@@ -6,6 +6,7 @@ import subprocess
 import sys
 import traceback
 import warnings
+from dataclasses import dataclass, field
 from datetime import datetime, timedelta
 from itertools import chain
 from logging import getLogger
@@ -14,18 +15,25 @@ from sys import stdout
 from time import sleep
 from typing import Optional, Iterable
 
-import ipynbname
 import pandas as pd
 from tabulate import tabulate
 
 from chrisbase.time import from_timestamp, now, str_delta
-from chrisbase.util import to_dataframe, tupled, SP, NO, OX
+from chrisbase.util import tupled, SP, NO, OX
 from chrisdict import AttrDict
 
 sys_stdout = sys.stdout
 sys_stderr = sys.stderr
 
 
+def cwd(path=None) -> Path:
+    if not path:
+        return Path.cwd()
+    else:
+        os.chdir(path)
+        return Path(path)
+
+
 def get_call_stack():
     def deeper():
         return list(traceback.walk_stack(None))
@@ -36,16 +44,32 @@ def get_call_stack():
     )
 
 
-def get_ipynb_path():
-    return ipynbname.path()
+# https://stackoverflow.com/questions/15411967/how-can-i-check-if-code-is-executed-in-the-ipython-notebook
+def is_notebook() -> bool:
+    try:
+        from IPython.core.getipython import get_ipython
+        shell = get_ipython().__class__.__name__
+        if shell == 'ZMQInteractiveShell':
+            return True  # Jupyter notebook or qtconsole
+        elif shell == 'TerminalInteractiveShell':
+            return False  # Terminal running IPython
+        else:
+            return False  # Other type
+    except NameError:
+        return False
 
 
-def get_current_path():
-    call_stack = get_call_stack()
-    if any('run_code' in x['name'] for x in call_stack) or any('cell line' in x['name'] for x in call_stack):
-        return Path(get_ipynb_path())
+def running_file(known_path: Path or str = None):
+    if known_path and exists_or(Path(known_path)):
+        return Path(known_path)
+    elif is_notebook():
+        import ipynbname
+        return ipynbname.path()
     else:
-        return Path(call_stack[-1]['file'])
+        for call_stack in get_call_stack():
+            if call_stack['name'] == '<module>':
+                return Path(call_stack['file'])
+        raise RuntimeError("Cannot find current path")
 
 
 def hr(c="=", w=137, t=0, b=0):
@@ -164,7 +188,7 @@ class MuteStd:
             exit(22)
 
 
-class MyTimer:
+class JobTimer:
     def __init__(self, name=None, prefix=None, postfix=None, verbose=False, mt=0, mb=0, pt=0, pb=0, rt=0, rb=0, rc='-',
                  file=stdout, flush_sec=None, mute_logger=None, mute_warning=None):
         self.mute = open(os.devnull, 'w')
@@ -269,15 +293,15 @@ def exists_or(path):
 
 def first_path_or(path):
     try:
-        return paths(path)[0]
-    except IndexError:
+        return next(iter(paths(path)))
+    except StopIteration:
         return None
 
 
 def first_or(xs):
     try:
         return next(iter(xs))
-    except IndexError:
+    except StopIteration:
         return None
 
 
@@ -315,6 +339,7 @@ def glob_files(path, glob: str):
 
 
 def paths_info(*xs, to_pathlist=paths, to_filename=str, sort_key=None):
+    from chrisbase.util import to_dataframe
     records = []
     all_paths = []
     for path in xs:
@@ -418,12 +443,14 @@ def remove_dir(path, ignore_errors=False) -> Path:
     return path
 
 
-def remove_dir_check(path, ignore_errors=True, file=None):
+def remove_dir_check(path, real=True, verbose=False, ignore_errors=True, file=None):
     path = Path(path)
-    print(f"- {str(path):<40}: {OX(path.exists())}", end='', file=file)
-    if path.exists():
+    if verbose:
+        print(f"- {str(path):<40}: {OX(path.exists())}", end='', file=file)
+    if path.exists() and real:
         shutil.rmtree(path, ignore_errors=ignore_errors)
-    print(f" -> {OX(path.exists())}", file=file)
+    if verbose:
+        print(f" -> {OX(path.exists())}", file=file)
     return not path.exists()
 
 
@@ -438,7 +465,7 @@ def remove_any(path, sleep_sec=0.0) -> Path:
     return path
 
 
-def load_json(path, **kwargs) -> dict:
+def load_json(path: str | Path, **kwargs) -> dict:
     file = Path(path)
     assert file.exists() and file.is_file(), f"file={file}"
     try:
@@ -449,7 +476,7 @@ def load_json(path, **kwargs) -> dict:
         raise RuntimeError(f"Please validate json file!\n- path: {path}\n- type: {type(e).__qualname__}\n- detail: {e}")
 
 
-def save_json(obj, path, **kwargs):
+def save_json(obj: dict, path: str | Path, **kwargs):
     file = make_parent_dir(Path(path))
     with file.open("w") as f:
         json.dump(obj, f, **kwargs)
@@ -565,12 +592,20 @@ def save_rows(rows, file, open_mode='w', keys=None, excl=None, with_column_name=
 
 
 def run_command(*args, title=None, mt=0, mb=0, pt=0, pb=0, rt=0, rb=0, rc='-', bare=False, verbose=True, real=True):
-    with MyTimer(name=None if bare else f"run_command({title})" if title else f"run_command{args}",
-                 verbose=verbose, mt=mt, mb=mb, pt=pt, pb=pb, rt=rt, rb=rb, rc=rc) as scope:
+    with JobTimer(name=None if bare else f"run_command({title})" if title else f"run_command{args}",
+                  verbose=verbose, mt=mt, mb=mb, pt=pt, pb=pb, rt=rt, rb=rb, rc=rc) as scope:
         if real:
             subprocess.run(list(map(str, args)), stdout=None if verbose else scope.mute, stderr=None if verbose else scope.mute)
 
 
+def read_command_out(*args):
+    return subprocess.run(list(map(str, args)), stdout=subprocess.PIPE).stdout.decode('utf-8')
+
+
+def read_command_err(*args):
+    return subprocess.run(list(map(str, args)), stderr=subprocess.PIPE).stderr.decode('utf-8')
+
+
 def get_valid_lines(lines,
                     accumulating_querys=(
                             ("(Epoch ", "training #1"),
@@ -599,27 +634,43 @@ def trim_output(infile, outfile):
     outfile.write_text('\n'.join(get_valid_lines(all_lines(infile))))
 
 
-def get_hostname():
+def get_hostname() -> str:
     return socket.gethostname()
 
 
-def get_hostaddr():
-    return socket.gethostbyname(get_hostname())
+def get_hostaddr(default="127.0.0.1") -> str:
+    with socket.socket(socket.AF_INET, socket.SOCK_DGRAM) as st:
+        st.connect(("8.8.8.8", 80))
+        r = first_or(st.getsockname())
+        return r if r else default
 
 
-def get_python_path():
-    return subprocess.run("which python".split(), stdout=subprocess.PIPE).stdout.decode('utf-8').strip()
+def prepend_to_global_path(*xs):
+    os.environ['PATH'] = os.pathsep.join(map(str, xs)) + os.pathsep + os.environ['PATH']
 
 
-def set_cuda_path(cuda_paths=("/usr/local/cuda-12.0", "/usr/local/cuda-11.4", "/usr/local/cuda-11.3", "/usr/local/cuda-11.1",
-                              "/usr/local/cuda-12", "/usr/local/cuda-11", "/usr/local/cuda")):
+def append_to_global_path(*xs):
+    os.environ['PATH'] = os.environ['PATH'] + os.pathsep + os.pathsep.join(map(str, xs))
+
+
+def include_cuda_bin_dir(candidate_dirs=None) -> Path:
+    if candidate_dirs is None:
+        candidate_dirs = sorted(dirs("/usr/local/cuda*"), reverse=True)
     cuda_dir = None
-    for cuda_path in cuda_paths:
-        if exists_or(cuda_path):
-            cuda_dir = cuda_path
+    for candidate_dir in candidate_dirs:
+        if exists_or(candidate_dir) and exists_or(f"{candidate_dir}/bin"):
+            cuda_dir = Path(candidate_dir)
             break
-    assert cuda_dir is not None
-    os.environ['PATH'] = f"{cuda_dir}/bin:{os.environ['PATH']}"
+    if cuda_dir:
+        prepend_to_global_path(f"{cuda_dir}/bin")
+    return cuda_dir
+
+
+def working_gpus(gpus=None):
+    if gpus:
+        os.environ["CUDA_DEVICE_ORDER"] = "PCI_BUS_ID"
+        os.environ["CUDA_VISIBLE_DEVICES"] = gpus
+    return os.environ.get("CUDA_VISIBLE_DEVICES")
 
 
 def set_torch_ext_path(dev=1):
@@ -629,3 +680,29 @@ def set_torch_ext_path(dev=1):
 
 def set_tokenizers_parallelism(value=False):
     os.environ["TOKENIZERS_PARALLELISM"] = f"{value}".lower()
+
+
+def environ_to_dataframe(max_value_len=200, columns=None):
+    from chrisbase.util import to_dataframe
+    return to_dataframe(copy_dict(dict(os.environ),
+                                  keys=[x for x in sorted(os.environ.keys()) if len(str(os.environ[x])) <= max_value_len]),
+                        columns=columns)
+
+
+@dataclass
+class BasicProjectEnv:
+    hostname: str = field(init=False)
+    hostaddr: str = field(init=False)
+    python_path: Path = field(init=False)
+    project_name: str = field()
+    project_path: Path = field(init=False)
+    working_path: Path = field(init=False)
+    running_file: Path = field(init=False)
+
+    def __post_init__(self):
+        self.hostname = get_hostname()
+        self.hostaddr = get_hostaddr()
+        self.python_path = Path(sys.executable)
+        self.project_path = cwd(first_or([x for x in running_file().parents if self.project_name and x.name.startswith(self.project_name)]))
+        self.working_path = Path.cwd()
+        self.running_file = running_file().relative_to(self.working_path)
